import tempfile
from pathlib import Path
from typing import Any, Dict, cast
from unittest import TestCase
from unittest.mock import Mock, call, patch

from epplib.commands.base import Login, Logout
from epplib.exceptions import TransportError
from testfixtures import LogCapture

from eppic.epp import EppHandler, _noop
from eppic.exceptions import EppicError, ServerError
from eppic.settings import Session

from .utils import DummyEppCommand as DummyRequest


@patch("eppic.epp.Client")
@patch("eppic.epp.SocketTransport")
class TestEppHandler(TestCase):
    def setUp(self) -> None:
        self.temp_dir = tempfile.TemporaryDirectory()
        keys = Path(self.temp_dir.name)
        with open(keys / "key", "w") as f:
            f.write("foobar")
        with open(keys / "cert", "w") as f:
            f.write("foobar")

        self.init: Dict[str, Any] = {
            "hostname": "hello",
            "port": 123,
            "cert_file": keys / "cert",
            "key_file": keys / "key",
            "username": "foo",
            "password": "bar",
            "obj_uris": ["foo", "bar"]
        }

    @property
    def session(self) -> Session:
        """Generate session."""
        return Session(**self.init)

    def tearDown(self) -> None:
        self.temp_dir.cleanup()

    def test_set_session(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(None, _noop)
        epp.session = self.session
        self.assertTrue(epp.logged_in)
        self.assertEqual(epp.session.username, self.init["username"])
        self.assertEqual(epp.session.hostname, self.init["hostname"])
        self.assertIsInstance(cast(Mock, epp.client.send).call_args_list[0].args[0], Login)

    def test_login_invalid_connection(self, socket_transport, client):
        epp = EppHandler(None, _noop)
        client.return_value.send.side_effect = Exception('Gazpacho!')
        with LogCapture() as log:
            with self.assertRaisesRegex(Exception, 'Gazpacho!'):
                epp.session = self.session
        log.check_present(("eppic.epp", "WARNING", "Logout failed."))
        self.assertFalse(epp.logged_in)
        self.assertIsInstance(cast(Mock, epp.client.send).call_args_list[0].args[0], Login)

    def test_login_invalid_unknown_err(self, socket_transport, client):
        client.return_value.send.return_value.code = 2345
        epp = EppHandler(None, _noop)
        with self.assertRaisesRegex(ServerError, "Login was unsuccessful."):
            epp.session = self.session
        self.assertFalse(epp.logged_in)
        self.assertIsInstance(cast(Mock, epp.client.send).call_args_list[0].args[0], Login)

    def test_login_invalid_auth(self, socket_transport, client):
        client.return_value.send.return_value.code = 2501
        epp = EppHandler(None, _noop)
        with self.assertRaisesRegex(ServerError, "Invalid username or password."):
            epp.session = self.session
        self.assertFalse(epp.logged_in)
        self.assertIsInstance(cast(Mock, epp.client.send).call_args_list[0].args[0], Login)

    def test_logout(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(self.session, _noop)
        cast(Mock, epp.client).close = Mock()
        epp.logout()
        self.assertIsNone(epp._client)
        self.assertIsInstance(cast(Mock, epp.client.send).call_args_list[-1].args[0], Logout)
        cast(Mock, epp.client.close).assert_called_once_with()

    def test_logout_not_logged(self, socket_transport, client):
        self.init.pop("username")
        self.init.pop("password")
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(self.session, _noop)
        epp.logout()
        self.assertFalse(len(cast(Mock, epp.client.send).call_args_list))
        cast(Mock, epp.client.close).assert_not_called()
        cast(Mock, epp.client.connect).assert_not_called()

    def test_logout_error(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(self.session, _noop)
        client.return_value.close.side_effect = ValueError("foo")
        with LogCapture() as log:
            epp.logout()  # exception is not thrown.
        log.check_present(("eppic.epp", "WARNING", "Connection failed on close."))
        self.assertFalse(epp.logged_in)

    def test_send_request_set_cltrid(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(self.session, _noop)
        cast(Mock, epp.client).close = Mock()
        epp.send_request(DummyRequest("foobar"), "baz")
        self.assertIsNotNone(epp._client)
        self.assertIsInstance(cast(Mock, epp.client.send).call_args_list[-1].args[0], DummyRequest)
        self.assertEqual(cast(Mock, epp.client.send).call_args_list[-1].kwargs["tr_id"], "baz")

    def test_send_request_error(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        print_xml_mock = Mock()
        epp = EppHandler(self.session, print_xml_mock)
        # reset after the login

        cast(Mock, epp.client).close = Mock()
        cast(Mock, epp.client).send.side_effect = TransportError("foo")
        with LogCapture() as log:
            self.assertRaises(ServerError, epp.send_request, Mock())
        self.assertIsNone(epp._client)
        self.assertIsInstance(cast(Mock, epp.client.send).call_args_list[-1].args[0], Logout)
        log.check_present(("eppic.epp", "WARNING", "Logout failed."))
        print_xml_mock.assert_called_with(epp._client.last_raw_command)  # type: ignore[union-attr]
        cast(Mock, epp.client.close).assert_called_once_with()

    def test_set_session_again(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(self.session, _noop)
        session = self.session
        session.username = "foobarbaz"
        epp.session = session

        self.assertTrue(epp.logged_in)
        self.assertEqual(epp.session.username, "foobarbaz")
        self.assertEqual(epp.session.hostname, self.init["hostname"])

    def test_client_no_session(self, socket_transport, client):
        epp = EppHandler(None, _noop)
        self.assertRaises(EppicError, lambda: epp.client)

    def test_client_no_hostname(self, socket_transport, client):
        self.init.pop("hostname")
        self.init.pop("username")
        epp = EppHandler(self.session, _noop)
        self.assertRaisesRegex(EppicError, "Hostname not specified.", lambda: epp.client)

    def test_no_session(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(None, _noop)
        self.assertRaises(EppicError, lambda: epp.session)

    def test_disclose(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(self.session, _noop)
        self.assertIsNotNone(epp.policy)

    def test_print_raw_xml(self, socket_transport, client):
        client.return_value.send.return_value.code = 1000
        epp = EppHandler(self.session, epp_handler_print_xml := Mock())
        epp.send_request(DummyRequest("foobar"), "baz")
        self.assertListEqual(epp_handler_print_xml.mock_calls, [
            call(epp._client.last_raw_command),  # type: ignore
            call(epp._client.last_raw_response),  # type: ignore
            call(epp._client.last_raw_command),  # type: ignore
            call(epp._client.last_raw_response),  # type: ignore
        ])
